#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <time.h>
#include <unistd.h>
#include "LEU_LIB.h"

#define RUN_TIME 30 // max. running time [sec.]
int event_flag = 0;

void get_errstatus();

////////////////////////////////////////////////////
//  Basic controls
////////////////////////////////////////////////////

int connect(char *serial)
{
	int LE;

	// without serial #, auto connect //
	if(serial == NULL){
		printf("\n connect without serial\n");
		if( (LE = LEU_ConnectNoSerial(1024*1024)) != LEU_SUCCESS ){
			printf("\n fail! \n");
			return -1;
		}
	return LE;
	}

	// connection with serial # //
	printf("\n connect %s \n", serial);
	if( (LE = LEU_Connect(serial, 1024*1024)) != LEU_SUCCESS ){
		printf("\n fail! \n");
		return -1;
	}
	return LE;
}

int disconnect(int LE)
{
	if( LEU_Disconnect(LE) != LEU_SUCCESS ){
		return -1;
	}
	return 0;
}

int stop(int LE)
{
	if( LEU_Stop(LE) != LEU_SUCCESS ){
		printf("stop failed\n");
		get_errstatus();
		return -1;
	}
	return 0;
}

/// printout error codes //
void get_errstatus(){
	int error;
	LEU_GetLastError(&error);
	printf("Error code : %d \n", error);
	switch(error){
	case LEU_ERROR_NOINIT:
		printf("have not initialized\n");
		break;
	case LEU_ERROR_NOCONNECT:
		printf("not connected\n");
		break;
	case LEU_ERROR_MEASURING:
		printf("now measuring\n");
		break;
	case LEU_ERROR_PARAMETER:
		printf("parameter error\n");
		break;
	case LEU_ERROR_NOSERIAL:
		printf("no device has such a serial #\n");
		break;
	case LEU_ERROR_COM:
		printf("connection error\n");
		break;
	case LEU_ERROR_ALLOC:
		printf("buffer allocation error\n");
		break;
	case LEU_ERROR_TIMEOUT:
		printf("process time-out\n");
		break;
	case LEU_ERROR_VERSION:
		printf("too old device version\n");
		break;
	case LEU_ERROR_NODATA:
		printf("no data\n");
		break;
	case LEU_NOERROR:
	default:
		printf("unknown error\n");
		break;
	}
}

////////////////////////////////////////////////////
//  trigger setting
////////////////////////////////////////////////////

int set_mytrig(int LE, int len, uint8_t * trgdata, int len2, uint8_t * trgdata2)
{
	int i;
	LEU_TRIGGER trg;
	memset(&trg,0,sizeof(LEU_TRIGGER));

	trg.post_size = LEU_TRIGGER_SIZE_0; // stop immediately
	
	if(len>0 && len<9 && trgdata){
		trg.trig[0].action = 0x0; // first trigger just make a flag
		trg.trig[0].data_length = len;
		for(i=0; i<len; i++){
			trg.trig[0].data[i] = trgdata[i]; // trigger pattern
			trg.trig[0].mask[i] = 0xff; // mask (ff = active)
		}
	}
	if(len2>0 && len2<9 && trgdata2){
		// 2nd trig. stop the measurement
		trg.trig[1].action = LEU_TRIGGER_MEASURE_STOP 
			| LEU_TRIGGER_LAST_SEQUENCE; 
		trg.trig[1].data_length = len2;
		for(i=0; i<len2; i++){
			trg.trig[1].data[i] = trgdata2[i]; // trigger pattern
			trg.trig[1].mask[i] = 0xff; // mask (ff = active)
		}
	}

	if( LEU_SetTrigger(LE,trg) != LEU_SUCCESS ){
		printf("\n fail! \n");
		get_errstatus();
		return -1;
	}
	
	return 0;
}

////////////////////////////////////////////////////
//  Data read & User analysis
////////////////////////////////////////////////////

int read_data(int LE)
{
	LEU_data data;

	/* counters */
	static int c_hs = 0;
	static int c_fs = 0;
	static int c_ls = 0;
	static int c_sof = 0;
	static int c_ping = 0;

	static int c_setup = 0;
	static int c_in = 0;
	static int c_out = 0;

	static int c_ad_0 = 0;
	static int c_ad_1 = 0;
	static int c_ad_2 = 0;
	static int c_ad_3 = 0;

	static int c_ed_0 = 0;
	static int c_ed_1 = 0;
	static int c_ed_2 = 0;
	static int c_ed_3 = 0;

	static int c_ack = 0;
	static int c_nak = 0;
	static int c_stall = 0;
	static int c_nyet = 0;

	static int c_setup_ack = 0;
	static int c_in_ack = 0;
	static int c_out_ack = 0;

	static int c_setup_nak = 0;
	static int c_in_nak = 0;
	static int c_out_nak = 0;

	static int c_endp_0 = 0;
	static int c_endp_1 = 0;
	static int c_endp_2 = 0;
	static int c_endp_3 = 0;

	static int c_addr_0 = 0;
	static int c_addr_1 = 0;
	static int c_addr_2 = 0;
	static int c_addr_3 = 0;

	static int pid_tok = -1;
	static uint8_t endp = 255;
	static uint8_t addr = 255;

	// Read data from the buffer
	while( LEU_Read(LE, &data) != LEU_FAIL ){

		// notify special events
		switch(data.status){	
		case LEU_DATA_STATUS_DISCONNECT:
			event_flag |= 0x1;
			break; 
		case LEU_DATA_STATUS_TRIGGER_EXT:
			event_flag |= 0x2;
			break;
		case LEU_DATA_STATUS_TRIGGER:
			if(event_flag & 0x4) event_flag |= 0x8;
			event_flag |= 0x4;
			break;
		default: break;
		}
		
		if(data.packet.PID == LEU_DATA_PID_IN ||
		   data.packet.PID == LEU_DATA_PID_OUT ||
		   data.packet.PID == LEU_DATA_PID_SETUP )
			{
				// start of normal transaction

				// reset
				addr = 255;
				endp = 255;
				pid_tok = -1;

				// speeds
				switch(data.status){
				case LEU_DATA_STATUS_PACKET_LS:	c_ls ++; break;
				case LEU_DATA_STATUS_PACKET_FS:	c_fs ++; break;
				case LEU_DATA_STATUS_PACKET_HS:	c_hs ++; break;
				}
				// get endpoint & address
				if(LEUD_GetAddrEndp(&data, &addr, &endp) == LEU_FAIL){
					printf("Cannot get Address/Endpoint : PID error ?\n");
				}
				
				// count 0-3 for simple display
				switch(endp){
				case 0: c_endp_0 ++; break;
				case 1: c_endp_1 ++; break;
				case 2: c_endp_2 ++; break;
				case 3: c_endp_3 ++; break;
				default: break;
				}
				switch(addr){
				case 0: c_addr_0 ++; break;
				case 1: c_addr_1 ++; break;
				case 2: c_addr_2 ++; break;
				case 3: c_addr_3 ++; break;
				default: break;
				}
			}
		
		// tokens //
		switch(data.packet.PID){
		case LEU_DATA_PID_SETUP:
			pid_tok = 0;
			c_setup += 1;
			break;
		case LEU_DATA_PID_IN:
			pid_tok = 1;
			c_in += 1;
			break;
		case LEU_DATA_PID_OUT:
			pid_tok = 2;
			c_out += 1;
			break;
		case LEU_DATA_PID_SOF: c_sof ++; pid_tok = -1; break;
		case LEU_DATA_PID_PING: c_ping ++; pid_tok = -1; break;
		default: break;
		}

		// handshakes
		if(pid_tok > -1){ // alrady token defined
			switch(data.packet.PID){
			case LEU_DATA_PID_ACK:
				c_ack++;
				if(pid_tok == 0) c_setup_ack ++;
				else if(pid_tok == 1) c_in_ack ++;
				else if(pid_tok == 2) c_out_ack ++;
				pid_tok = -1;
				break;
			case LEU_DATA_PID_NAK:
				c_nak++;
				if(pid_tok == 0) c_setup_nak ++;
				else if(pid_tok == 1) c_in_nak ++;
				else if(pid_tok == 2) c_out_nak ++;
				pid_tok = -1;
				break;
			case LEU_DATA_PID_NYET:
				c_nyet++;
				pid_tok = -1;
				break;
			case LEU_DATA_PID_STALL:
				c_stall++;
				pid_tok = -1;
				break;
			}
		}
	}
	
	// clear console //
	printf("\033[2J\033[2;1H");
	// outputs //
	printf("High: %6d\n", c_hs);
	printf("Full: %6d\n", c_fs);
	printf("Low : %6d\n\n", c_ls);
	printf("SETUP : %6d / %6d ACK / %6d NAK\n", c_setup, c_setup_ack, c_setup_nak);
	printf("IN    : %6d / %6d ACK / %6d NAK\n", c_in, c_in_ack, c_in_nak);
	printf("OUT   : %6d / %6d ACK / %6d NAK\n\n", c_out, c_out_ack, c_out_nak);
	printf("ACK  : %6d\n", c_ack);
	printf("NAK  : %6d\n", c_nak);
	printf("NYET : %6d\n", c_nyet);
	printf("STALL: %6d\n\n", c_stall);
	printf("SOF : %6d\n",c_sof);
	printf("PING: %6d\n\n",c_ping);

	printf("ADDRESS  : 0 = %6d / 1 = %6d / 2 = %6d / 3 = %6d\n",
		   c_addr_0, c_addr_1, c_addr_2, c_addr_3);
	printf("ENDPOINT : 0 = %6d / 1 = %6d / 2 = %6d / 3 = %6d\n\n",
		   c_endp_0, c_endp_1, c_endp_2, c_endp_3);

	if(event_flag & 0x1){
		printf("!!! DISCONNECTED !!!\n");
		event_flag &= ~(0x1); // just once
	}
	if(event_flag & 0x2){
		printf("!!! EX-TRIGGER !!!\n");
		event_flag &= ~(0x2); // just once
	}
	
	if(event_flag & 0x4) printf("!!! TRIGGER1 DETECTED !!!\n");
	return 0;
}


int measure(int LE, uint8_t *trg, uint8_t *trg2)
{
	int i;
	int st;
	
	if( LEU_Measure(LE) != LEU_SUCCESS ){
		printf("measurement start failed\n");
		get_errstatus();
		return -1;
	}

	for(i=0; i<RUN_TIME; i++){
		sleep(1);
		read_data(LE);
		printf("waiting trigger = %s\n", (event_flag & 0x4) ? trg2 : trg);
		printf("%02d / %d sec. elapsed\n", i+1, RUN_TIME);
		
		if(LEU_GetState(LE, &st) == LEU_FAIL){
			printf("Cannot obtain device status\n");
			get_errstatus();
			break;
		}
		if(st != LEU_STATE_MEASURING){
			printf("Measurement stopped");
			if(event_flag & 0x8){printf(" by Trigger");}
			printf("\n");
			break;
		}
	}
	printf("\n");
	
	if(i == RUN_TIME)
		printf("Measurement stopped by timeout\n");
	stop(LE);
	
	return 0;
}


int main(int argc, char *argv[])
{
	int LE = -1;

	if( LEU_Init() != 0 ){
		printf("Init failed\n");
		get_errstatus();
		return 0;
	}

	// auto connect //
	//!!do not use with multiple analyzers!!//
	if(argc < 2) LE = connect(NULL);
	// normal connect //
	else LE = connect(argv[1]);

	if(LE == LEU_FAIL){
		get_errstatus();
		return -1;
	}
	uint8_t sdata[9], sdata2[9];
	int len = 0, len2 = 0;
	
	printf("\nplease enter the first data pattern to search (8 bytes max.)\n");
	fgets(sdata, sizeof(sdata), stdin);
	while( sdata[0] == 0x0a){
		getchar();
		fgets(sdata, sizeof(sdata), stdin);
	}
	
	len = strlen(sdata);
	if(len > 0 ){
		len --;
		sdata[len] = '\0'; // without \n
	}
	printf("1st trigger string set to \"%s\" (without newline codes)\n",sdata);

	printf("\nplease enter the second data pattern to search (8 bytes max.)\n");
	fgets(sdata2, sizeof(sdata2), stdin);
	while( sdata2[0] == 0x0a){
		getchar();
		fgets(sdata2, sizeof(sdata2), stdin);
	}
	
	len2 = strlen(sdata2);
	if(len2 > 0 ){
		len2 --;
		sdata2[len2] = '\0'; // without \n
	}
	printf("stop trigger string set to \"%s\" (without newline codes)\n",sdata2);

	printf("registering triggers..");
	fflush(stdout);
	
	if(set_mytrig(LE, len, sdata, len2, sdata2) == -1){
		return -1;
	}
	printf("done.\n");

	// hardware filter if needed //
	if(LEU_SetFilter(LE, 0x0) == -1){
		return -1;
	};
	/*
	if(LEU_SetFilter(LE, LEU_FILTER_IN_NAK | LEU_FILTER_SOF) == -1){
		return -1;
	};
	*/
	
	///////////// measurement ///////////////

	measure(LE, sdata, sdata2);

	// quit //
	disconnect(LE);
	LEU_End();
	
	return 0;
}
